use clap::Parser;
use log::debug;
use std::path::PathBuf;

use snafu::{prelude::*, Snafu};

#[derive(Debug, Snafu)]
pub enum Error {
    #[snafu(display("Io Error for \"{path}\": {source}"))]
    Io {
        path: String,
        source: std::io::Error,
    },
    #[snafu(display("Logger error: {source}"))]
    Logger { source: log::SetLoggerError },
}

#[derive(Parser)]
#[clap(author, version, about)]
pub struct Cli {
    /// Custom path to database
    #[clap(long = "database", value_name = "path", value_parser, default_value_t = database_path().into_os_string().into_string().unwrap())]
    pub database_path: String,
    /// Custom path to configuration
    #[clap(long = "config", value_name = "path", value_parser, default_value_t = config_path().into_os_string().into_string().unwrap())]
    pub config_path: String,
    /// Custom path to URLs with RSS feeds being fetched
    #[clap(long = "urls", value_name = "path", value_parser, default_value_t = urls_path().into_os_string().into_string().unwrap())]
    pub urls_path: String,
    /// Don't send emails
    #[clap(long, value_parser)]
    pub dry_run: bool,
    /// Don't fetch feeds
    #[clap(long, value_parser)]
    pub no_fetch: bool,
    /// Be verbose. Pass multiple times to increase verbosity.
    #[clap(short, action = clap::ArgAction::Count)]
    pub verbose: u8,
    /// Email "From" header
    #[clap(long, value_name = "from", value_parser)]
    pub mail_from: Option<String>,
    /// Email "To" header
    #[clap(long, value_name = "to", value_parser)]
    pub mail_to: Option<String>,
    /// SMTP user
    #[clap(long, value_name = "user", value_parser)]
    pub smtp_user: Option<String>,
    /// SMTP password
    #[clap(long, value_name = "password", value_parser)]
    pub smtp_password: Option<String>,
    /// SMTP server
    #[clap(long, value_name = "server", value_parser)]
    pub smtp_server: Option<String>,
    /// SMTP server port
    #[clap(long, value_name = "port", value_parser = clap::value_parser!(u16).range(1..))]
    pub smtp_port: Option<u16>,
}

impl Cli {
    /// Parse the clap `Cli` struct with the command line arguments and create the database if it does not exist yet.
    pub fn build_app() -> Result<Self, Error> {
        use std::fs::File;

        let args = Self::parse();

        // setup logging as soon as possible
        let verbosity = match args.verbose {
            0 => log::Level::Warn,
            1 => log::Level::Info,
            2 => log::Level::Debug,
            _ => log::Level::Trace,
        };
        simple_logger::init_with_level(verbosity).context(LoggerSnafu)?;

        let data_dir = data_dir();
        if !PathBuf::from(&data_dir).exists() {
            debug!(
                "No data directory exists, creating a new one: {:?}",
                &data_dir
            );
            std::fs::create_dir(&data_dir).with_context(|_| IoSnafu {
                path: data_dir.display().to_string(),
            })?;
        }

        if File::open(&args.database_path).is_err() {
            debug!(
                "No database file exists, creating a new one: {:?}",
                &args.database_path
            );
            File::create(&args.database_path).with_context(|_| IoSnafu {
                path: args.database_path.clone(),
            })?;
        }

        Ok(args)
    }
}

fn project_dirs() -> directories::ProjectDirs {
    #[allow(clippy::unwrap_used)]
    directories::ProjectDirs::from("", "", "rss-email")
        .expect("A valid home directory path from the OS")
}

/// Returns the base directory where all the configuration files are stored.
fn config_dir() -> PathBuf {
    project_dirs().config_dir().to_path_buf()
}

/// Returns the base directory where all the data files are stored.
fn data_dir() -> PathBuf {
    project_dirs().data_dir().to_path_buf()
}

/// Returns the path to the database.
#[allow(dead_code)]
fn database_path() -> PathBuf {
    data_dir().join("cache.db")
}

/// Returns the path to the configuration file.
#[allow(dead_code)]
fn config_path() -> PathBuf {
    config_dir().join("config.toml")
}

/// Returns the path to the urls file.
#[allow(dead_code)]
fn urls_path() -> PathBuf {
    config_dir().join("urls")
}
