use serde::Deserialize;
use snafu::{prelude::*, Snafu};
use std::fs::File;
use std::{io::Read, path::Path};

#[derive(Debug, Snafu)]
pub enum Error {
    #[snafu(display("Io Error for \"{path}\": {source}"))]
    Io {
        path: String,
        source: std::io::Error,
    },
    Toml {
        source: toml::de::Error,
    },
    #[snafu(display("Cli error: {source}"))]
    Cli {
        source: crate::cli::Error,
    },
}

#[derive(Debug, Deserialize)]
pub struct Config {
    pub(crate) mail: Mail,
    pub(crate) smtp: SmtpConfig,
}

#[derive(Debug, Deserialize)]
pub struct Mail {
    pub(crate) to: String,
    pub(crate) from: String,
}

#[derive(Debug, Deserialize)]
pub struct SmtpConfig {
    pub(crate) user: String,
    pub(crate) password: String,
    pub(crate) server: String,
    pub(crate) port: u16,
}

impl Config {
    pub fn new(config_path: impl AsRef<Path>) -> Result<Self, Error> {
        let config_path = config_path.as_ref();
        let mut string = String::new();
        File::open(config_path)
            .with_context(|_| IoSnafu {
                path: config_path.display().to_string(),
            })?
            .read_to_string(&mut string)
            .with_context(|_| IoSnafu {
                path: config_path.display().to_string(),
            })?;
        let config: Self = toml::de::from_str(&string).context(TomlSnafu)?;

        Ok(config)
    }
}

/// The global configuration for the whole program.
///
/// This is the global configuration struct. Instead of carrying around the command line arguments
/// and config file values, this struct holds both, overriding config file values with command line
/// ones.
#[derive(Debug, Default)]
pub struct AppConfig {
    pub database_path: String,
    pub urls_path: String,
    pub dry_run: bool,
    pub no_fetch: bool,
    pub mail_from: String,
    pub mail_to: String,
    pub smtp_user: String,
    pub smtp_password: String,
    pub smtp_server: String,
    pub smtp_port: u16,
}

impl AppConfig {
    /// Create a new [`AppConfig`].
    ///
    /// Internally, this will parse the command line arguments with [`clap`] and read the
    /// configuration file *if* all of the required options have *not* been given.
    pub fn new() -> Result<Self, Error> {
        let cli = crate::cli::Cli::build_app().context(CliSnafu)?;
        let mut s = Self {
            database_path: cli.database_path,
            urls_path: cli.urls_path,
            dry_run: cli.dry_run,
            no_fetch: cli.no_fetch,
            ..Default::default()
        };

        // if all the values on the command line that correspond to config file options have a
        // value given, there is no need to require a config file at all
        match (
            cli.mail_from,
            cli.mail_to,
            cli.smtp_user,
            cli.smtp_password,
            cli.smtp_server,
            cli.smtp_port,
        ) {
            (
                Some(mail_from),
                Some(mail_to),
                Some(smtp_user),
                Some(smtp_password),
                Some(smtp_server),
                Some(smtp_port),
            ) => {
                trace!("All necessary config values have been given on the command line, no need to check for config file");
                s.mail_from = mail_from;
                s.mail_to = mail_to;
                s.smtp_user = smtp_user;
                s.smtp_password = smtp_password;
                s.smtp_server = smtp_server;
                s.smtp_port = smtp_port;
            }
            (mail_from, mail_to, smtp_user, smtp_password, smtp_server, smtp_port) => {
                let cfg = Config::new(cli.config_path)?;
                s.mail_from = mail_from.unwrap_or(cfg.mail.from);
                s.mail_to = mail_to.unwrap_or(cfg.mail.to);
                s.smtp_user = smtp_user.unwrap_or(cfg.smtp.user);
                s.smtp_password = smtp_password.unwrap_or(cfg.smtp.password);
                s.smtp_server = smtp_server.unwrap_or(cfg.smtp.server);
                s.smtp_port = smtp_port.unwrap_or(cfg.smtp.port);
            }
        }

        Ok(s)
    }
}
