use atom_syndication;
use rss;
use snafu::{prelude::*, Snafu};

use crate::models::Post;

#[derive(Debug, Snafu)]
pub enum Error {
    #[snafu(display("reqwest got error fetching feed \"{url}\": {source}"))]
    ReceivingFeed { url: String, source: reqwest::Error },
    #[snafu(display("RSS feed is formatted incorrectly: {url}"))]
    InvalidRSS { url: String },
    #[snafu(display("Atom feed is formatted incorrectly: {url}"))]
    InvalidAtom { url: String },
    #[snafu(display("Feed \"{url}\" is neither an RSS nor an Atom feed"))]
    InvalidFeed { url: String },
}

/// Fetch a new feed, RSS or Atom.
///
/// # Errors
///
/// An error occurs if the feed cannot be fetched or is invalid.
pub async fn fetch_new<S>(url: S) -> Result<Vec<Post>, Error>
where
    S: AsRef<str> + Send,
{
    debug!("Fetching feed for {}", url.as_ref());
    let url = url.as_ref();
    let content = reqwest::get(url)
        .await
        .with_context(|_| ReceivingFeedSnafu {
            url: url.to_string(),
        })?
        .bytes()
        .await
        .with_context(|_| ReceivingFeedSnafu {
            url: url.to_string(),
        })?;
    match fetch_new_rss(&content[..]) {
        Err(rss::Error::InvalidStartTag) => match fetch_new_atom(&content[..]) {
            Err(atom_syndication::Error::InvalidStartTag) => InvalidFeedSnafu {
                url: url.to_string(),
            }
            .fail(),
            Err(_) => InvalidAtomSnafu {
                url: url.to_string(),
            }
            .fail(),
            Ok(v) => Ok(v),
        },
        Err(_) => InvalidRSSSnafu {
            url: url.to_string(),
        }
        .fail(),
        Ok(v) => Ok(v),
    }
}

/// Build a new rss feed from a given byte slice.
///
/// # Errors
///
/// An error occurs if the given bytes are an invalid RSS feed.
pub fn fetch_new_rss(bytes: &[u8]) -> Result<Vec<Post>, rss::Error> {
    let channel = rss::Channel::read_from(bytes)?;

    Ok(channel
        .items
        .into_iter()
        .filter_map(|i| match i.try_into() {
            Ok(p) => Some(p),
            Err(e) => {
                error!("Unable to convert received post, continuing ({e})");
                None
            }
        })
        .collect::<Vec<Post>>())
}

/// Build a new Atom feed from a given byte slice.
///
/// # Errors
///
/// An error occurs if the given bytes are an invalid Atom feed.
pub fn fetch_new_atom(bytes: &[u8]) -> Result<Vec<Post>, atom_syndication::Error> {
    let feed = atom_syndication::Feed::read_from(bytes)?;

    Ok(feed
        .entries
        .into_iter()
        .filter_map(|e| match e.try_into() {
            Ok(e) => Some(e),
            Err(e) => {
                error!("Unable to convert received post, continuing ({e})");
                None
            }
        })
        .collect::<Vec<_>>())
}
