use lettre::{
    message::Message,
    transport::smtp::{authentication::Credentials, AsyncSmtpTransport},
    AsyncTransport, Tokio1Executor,
};
use snafu::{prelude::*, Snafu};

#[derive(Debug, Snafu)]
pub enum Error {
    #[snafu(display("General Mail error: {source}"))]
    GeneralMail { source: lettre::error::Error },
    #[snafu(display("SMTP Mail error: {source}"))]
    SMTP {
        source: lettre::transport::smtp::Error,
    },
    #[snafu(display("Unable to parse {value}"))]
    Parse { value: String },
}

pub struct Mail {
    subject: String,
    body: String,
}

impl Mail {
    /// Build a Mail.
    ///
    /// Build a Mail from the optional components subject, body, and url.
    /// If any of the parameters are not given, they will be replaced with a default value.
    #[must_use]
    pub fn new(subject: Option<String>, body: Option<String>, url: Option<String>) -> Self {
        let subject = subject.unwrap_or_else(|| "No title found".to_string());
        let url = url.unwrap_or_else(|| "No url found".to_string());
        let body = match body {
            Some(c) => url + "\n\n" + &c,
            None => url,
        };

        Self { subject, body }
    }

    /// Send an email with the body of `self`.
    ///
    /// # Errors
    ///
    /// An error occurs if the email cannot be built or the email cannot be sent.
    pub async fn send_email<'a>(
        self,
        from: &'a str,
        to: &'a str,
        mailer: &AsyncSmtpTransport<Tokio1Executor>,
    ) -> Result<(), Error> {
        trace!("Sending to {}: {}", to, &self.subject);
        let from = from.parse().map_err(|_| Error::Parse {
            value: from.to_string(),
        })?;
        let to = to.parse().map_err(|_| Error::Parse {
            value: to.to_string(),
        })?;

        let email = Message::builder()
            .from(from)
            .to(to)
            .subject(self.subject)
            .body(self.body)
            .context(GeneralMailSnafu)?;

        mailer.send(email).await.context(SMTPSnafu)?;

        Ok(())
    }
}

/// Get the mailer for sending mail.
///
/// # Errors
///
/// An error occures if instantiating a [`AsyncSmtpTransport`] in the body fails
///
/// [`AsyncSmtpTransport`]: lettre::transport::smtp::AsyncSmtpTransport
pub fn get_mailer(
    user: String,
    password: String,
    server: &str,
    port: u16,
) -> Result<AsyncSmtpTransport<Tokio1Executor>, Error> {
    let creds = Credentials::new(user, password);

    let mailer = AsyncSmtpTransport::<Tokio1Executor>::relay(server)
        .context(SMTPSnafu)?
        .credentials(creds)
        .port(port)
        .build();

    Ok(mailer)
}
