#![deny(
    clippy::pedantic,
    clippy::nursery,
    clippy::unwrap_used,
    clippy::cargo,
    clippy::style,
    clippy::complexity,
    clippy::correctness,
    clippy::unseparated_literal_suffix,
    clippy::try_err,
    clippy::almost_swapped,
    clippy::approx_constant
)]
#![allow(clippy::missing_errors_doc, clippy::module_name_repetitions)]

#[macro_use]
extern crate log;

pub mod cli;
pub mod config;
pub mod db;
pub mod feed;
pub mod mail;
pub mod models;

use crate::{
    config::AppConfig,
    mail::{get_mailer, Mail},
};
use lettre::{AsyncSmtpTransport, Tokio1Executor};
use snafu::{prelude::*, Snafu};
use std::{
    fs::File,
    io::{BufRead, BufReader},
    sync::Arc,
};

use sqlx::{sqlite::SqlitePoolOptions, Sqlite};
use tokio::task::JoinSet;

#[derive(Debug, Snafu)]
pub enum Error {
    #[snafu(display("Io Error at \"{path}\": {source}"))]
    Io {
        path: String,
        source: std::io::Error,
    },
    #[snafu(display("General database error: {source}"))]
    GeneralDatabaseError { source: sqlx::Error },
    #[snafu(display("Migration error: {source}"))]
    MigrationError { source: sqlx::migrate::MigrateError },
    #[snafu(display("Task in JoinSet failed to execute to completion: {source}"))]
    JoinSet { source: tokio::task::JoinError },
    #[snafu(display("Mail error: {source}"))]
    Mail { source: crate::mail::Error },
    #[snafu(display("Config error: {source}"))]
    Config { source: crate::config::Error },
}

#[tokio::main]
async fn main() {
    if let Err(e) = app_main().await {
        eprintln!("{e}");
        std::process::exit(1);
    }
}

async fn app_main() -> Result<(), Error> {
    let config = Arc::new(AppConfig::new().context(ConfigSnafu)?);
    let urls: Vec<String> =
        BufReader::new(File::open(config.urls_path.as_str()).context(IoSnafu {
            path: &config.urls_path,
        })?)
        .lines()
        .map_while(Result::ok)
        .filter(|l| !l.starts_with('#'))
        .collect();

    let db_path = &config.database_path;
    debug!("Establishing connection to database at {:?}", db_path);
    let pool = SqlitePoolOptions::new()
        .max_connections(5)
        .connect(db_path.as_str())
        .await
        .context(GeneralDatabaseSnafu)?;

    sqlx::migrate!("./migrations")
        .run(&pool)
        .await
        .context(MigrationSnafu)?;

    if config.no_fetch {
        info!("Not fetching any feeds as \"--no-fetch\" has been passed");
    } else {
        fetch_feeds(urls, &pool).await?;
    }

    let results = sqlx::query_as!(
        models::Post,
        "select * from posts where sent != true order by pub_date desc"
    )
    .fetch_all(&pool)
    .await
    .context(GeneralDatabaseSnafu)?;

    let mailer = get_mailer(
        config.smtp_user.clone(),
        config.smtp_password.clone(),
        &config.smtp_server,
        config.smtp_port,
    )
    .context(MailSnafu)?;

    let mut handles = JoinSet::new();
    for result in results {
        let mut conn = pool.acquire().await.context(GeneralDatabaseSnafu)?;
        let mailer = mailer.clone();
        let config = config.clone();

        handles.spawn(async move {
            send_post(
                &mut conn,
                mailer,
                &config.mail_from,
                &config.mail_to,
                result,
                config.dry_run,
            )
            .await
        });
    }

    while let Some(handle) = handles.join_next().await {
        // TODO: retry sending mail instead? user should specify number of retries in config
        if let Err(e) = handle.context(JoinSetSnafu)? {
            log::error!("An error occured while sending an email: {e}");
        }
    }

    Ok(())
}

async fn fetch_feeds(urls: Vec<String>, pool: &sqlx::Pool<Sqlite>) -> Result<(), Error> {
    let mut set = JoinSet::new();
    for u in urls {
        set.spawn(async move { feed::fetch_new(u).await });
    }

    while let Some(new) = set.join_next().await {
        let posts = match new.context(JoinSetSnafu)? {
            Ok(p) => p,
            Err(e) => {
                log::error!("Error while fetching feed: {e}");
                continue;
            }
        };

        for i in posts {
            let conn = pool.acquire().await.context(GeneralDatabaseSnafu)?;
            db::insert_item(conn, &i).await?;
        }
    }

    Ok(())
}

async fn send_post<'a, E>(
    conn: E,
    mailer: AsyncSmtpTransport<Tokio1Executor>,
    from: &'a str,
    to: &'a str,
    post: models::Post,
    dry_run: bool,
) -> Result<(), Error>
where
    E: sqlx::Executor<'a, Database = Sqlite>,
{
    if dry_run {
        info!("Not sending any emails as \"--dry-run\" has been passed");
    } else {
        Mail::new(post.title, post.content, post.url)
            .send_email(from, to, &mailer)
            .await
            .context(MailSnafu)?;
    }

    sqlx::query!("update posts set sent = true where guid = ?", post.guid)
        .execute(conn)
        .await
        .context(GeneralDatabaseSnafu)?;

    Ok(())
}
