use chrono::DateTime;
use snafu::Snafu;

#[derive(Debug, Snafu)]
#[snafu(display("Missing GUID on item: {item:?}"))]
pub struct MissingGUID {
    item: rss::Item,
}

#[derive(Debug)]
pub struct Post {
    pub guid: String,
    pub title: Option<String>,
    pub url: Option<String>,
    pub pub_date: Option<i64>,
    pub content: Option<String>,
    pub sent: bool,
}

impl TryFrom<rss::Item> for Post {
    type Error = MissingGUID;

    fn try_from(item: rss::Item) -> Result<Self, Self::Error> {
        let time = item.pub_date().map(|date| {
            DateTime::parse_from_rfc2822(date)
                .unwrap_or_else(|_| DateTime::default())
                .timestamp()
        });

        let guid = item
            .guid()
            .ok_or_else(|| MissingGUID { item: item.clone() })?
            .value()
            .to_string();
        let title = item.title().map(String::from);
        let url = item.link().map(String::from);
        let pub_date = time;
        let content = item
            .content()
            .or_else(|| item.description())
            .map(String::from);

        Ok(Self {
            guid,
            title,
            url,
            pub_date,
            content,
            sent: false,
        })
    }
}

// `TryFrom` is implemented for consistency, not because it can actually fail at the moment
impl TryFrom<atom_syndication::Entry> for Post {
    // NOTE: NOT USED
    type Error = MissingGUID;

    fn try_from(value: atom_syndication::Entry) -> Result<Self, Self::Error> {
        let guid = value.id.clone();
        let title = Some(value.title.value);
        let url = Some(value.id);
        let pub_date = value.published.map(|p| p.timestamp());
        let content = if let Some(c) = value.content {
            c.value
        } else {
            None
        };

        Ok(Self {
            guid,
            title,
            url,
            pub_date,
            content,
            sent: false,
        })
    }
}
